/////////////////////////////////////////////////////////////////////////////
// Name:        Do178Builder.cpp 
// Author:      Ron Burkey
// Purpose:     This is sort of the top-level source file for the 
//              Do178Builder application.  It contains the code for the 
//              Do178BuilderApp and Do178BuilderFrame classes.
// Mods:        09/21/01 RSB    Created.  This file was created initially 
//                              by the wizard of the wxDesigner program,
//                              and is modified somewhat every time I use 
//                              wxDesigner to tweak the user interface. 
//                              In fact, wxDesigner puts all of the code
//                              it generates here, tending to make this file
//                              very large and complex.  What I've done,
//                              therefore, is to leave the code for the
//                              Do178BuilderApp and Do178BuilderFrame
//                              classes here, as well as all event tables.
//                              But each significant class, such as
//                              XmlEditorDialogClass is given its own source
//                              file (like XmlEditorDialogClass.cpp),
//                              and all of the code for those classes 
//                              (except event tables) is moved to those files.
//                              The program comments in this file beginning
//                              "WDR:" are used by wxDesigner, and should
//                              not be modified.
//              10/21/01 RSB    ... still under continual adjustment ...
//              02/17/02 RSB    Adjusted the splash screen, etc.
//                              Added automatic loading of XML from command
//                              line.
//              02/18/02 RSB    Added OnSize to the XmlEditorDialogClass.
//              04/08/02 RSB    The XmlEditorDialogClass::FindButton function
//                              did not compile with wxWindows 2.3.2, because
//                              the wxDynamicCast function apparently no longer
//                              works.  These have all been replaced by
//                              "getter" functions instead.
//              08/06/03 RSB    Now attempts to maximize upon opening the
//                              editor.  HOWEVER, refuses to expand beyond
//                              1280x1024.  The reason for this is that
//                              on my RedHat 7.3 system, expanding beyond
//                              this causes the program to intermittently
//                              segfault thereafter.  I have not been able
//                              to track down the ultimate cause of this,
//                              as it seems to have something to do with
//                              wxWindows or GTK or X, and not directly
//                              to do with Do178Builder.  Also, changed the way
//                              the -p command-line switch is used, so that 
//                              (if present) automatically does a chdir to that
//                              directory.  (The way directories are used 
//                              confuses me a little, so I'm not sure if this
//                              mightn't break something else.)
//              08/07/03 RSB    Added -x and -y switches.
//              11/10/03 RSB    The output directory name now includes the
//                              datestamp by default, but this can be 
//                              overridden with the -o switch.
//              11/11/03 RSB    Kludged a workaround for the fact that I can't
//                              seem to get wxClientDisplayRect in Win32 with
//                              wxGTK 2.7.
//              03/26/04 RSB    Updated copyright.
//              05/02/04 RSB    Added locking for SaveAs.  Also, added stuff
//                              to file-locking to try to recover from a 
//                              messed-up server that keeps stuff locked
//                              when nobody is using it.  Also, explicitly
//                              unlock on exiting the program.
//              06/20/05 RSB    Added --font-size.
//              05/15/06 RSB    Improved the error reporting slightly for 
//                              LoadXmlFile errors.
//              06/23/06 RSB    Began adapting for DO-254.
//              06/29/06 RSB    Pruning to an assurance level added.
//              07/02/06 RSB    Added read-only stuff.
//              07/04/06 RSB    Now builds with wxGTK-2.6.x and gcc 4.x.
//              07/06/06 RSB    Maximization on startup in Windows fixed.
//                              The main window and editor dialog titles
//                              now indicate whether DO-254 or DO-178B is
//                              active.  Added ResizeWorkaround.  (Verified
//                              that the problem of Do178Builder suddenly
//                              quitting after the dialog has been resized
//                              >1280x1024 still exists in RedHat 7.3, so
//                              the default is still to use the workaround.)
//              07/08/06 RSB    Corrected the status messages used when
//                              documents are built, so that the call out
//                              DO-254 stuff when appropriate.
//              03/12/07 RSB    Added the --no-forward-trace and 
//                              --no-trace switches.
//              01/22/08 RSB    Added the --combine-srd-sdd switch.
//              01/26/08 RSB    Added the --x45 switch.
//              01/27/08 RSB    Added the --x412, --x42, --x43, and --x44
//                              switches.
//              03/08/08 RSB    Added the --no-productname switch.
//              03/26/08 RSB    Accounted for the fact that wxTreeItemId
//                              is no longer an integer type object in 
//                              in wxWidgets 2.8 as it was in 2.6 and 
//                              prior, so that some of the assignments 
//                              that are being made to them are invalid
//                              in 2.8.
//              04/05/08 RSB    Restuctured the document-build function
//                              so that it can be activated from the 
//                              editor as well as from the menu.
//              01/27/09 RSB    Added the --batch command-line switch.
//		01/28/09 RSB	Fixed an enormous number of errors and/or warnings
//				that appear with newest version of g++ and/or
//				Unicode builds of wxWidgets.
//              02/03/09 RSB    Added the --combine-plans switch.

#define COPYRIGHT_YEARS "2001-2009"
/*
  Copyright 2001-2009 Ronald S. Burkey

  This file is part of Do178Builder.

  Do178Builder is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  Do178Builder is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Do178Builder; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/////////////////////////////////////////////////////////////////////////////

#ifdef __GNUG__
#pragma implementation "Do178Builder.cpp"
#endif

// For compilers that support precompilation
#include "wx/wxprec.h"

#ifdef __BORLANDC__
#pragma hdrstop
#endif

// Include private headers
#include "Do178Builder.h"
#include "Do178Builder_wdr.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <unistd.h>

// From LoadXmlFile.cpp.
void PruneToAssuranceLevel (struct XmlStruct *Root, int AssuranceLevel);
void NormalizeApplicability (struct XmlStruct *Root);

// Here are some things needed only in Win32
#ifndef NAME_MAX
#define NAME_MAX 512
#endif
#ifdef __WINDOWS__
#define MKDIR(x) mkdir(x)
#else
#define MKDIR(x) mkdir(x,0750)
#endif

int NoProductname = 0;
static int ResizeWorkaround = 1;
int NoForwardTrace = 0, NoTrace = 0;
int CombineSrdSdd = 0;
char *CombinedPlanName = NULL;
int x4 = 0;
const char *x4v1 = "x", *x4v2 = "x";
bool HintEditable = FALSE;
bool TemporaryHintEditable = FALSE;
int DatasetChanged = 0;
wxString rFilename, nFilename;
static char CurrentDirectory[NAME_MAX + 1] = ".";
static char HomeDirectory[NAME_MAX + 1] = ".";
static struct XmlStruct Xml = { NULL }, TemplateXml;
int CopyrightSplash = 1;
static char s[1024];		// A dummy array used several places.
static int MaxWindowX = 1280, MaxWindowY = 1024;
static int OutputOverride = 0;

int FontSize = 14, TreeFontSize = -12;
int Do254 = 0;
const char *TemplateFilename = ".template.xml";

// Stuff for the lock server, presumed to be Alberto Bertogli's 
// Open Lock Daemon ("old").
#ifdef WIN32
static const char *OldServer = "(none)";
#ifndef OLD_PORT
#define OLD_PORT 0
#endif //OLD_PORT
#else // WIN32
static const char *OldServer = "localhost";
#endif // WIN32
static int OldPort = OLD_PORT;
int OldFd = -1;

int ReadOnly = 0;

// This is just a unique item that isn't supposed to correspond to any
// item actually in our wxTreeCtrl.  As long as we don't fiddle with it
// other than constructing it, that's exactly what it should be.  For
// wxWidgets 2.6 and prior, this is an integer-type thing and so (I think)
// doesn't have a constructor.
#if wxMAJOR_VERSION == 2 && wxMINOR_VERSION <= 6
wxTreeItemId INVALID_TREE_ITEM = (wxTreeItemId) - 1;
#else
wxTreeItemId INVALID_TREE_ITEM;
#endif

// Normalize the filename, so that it is that same from every client.
static wxString Name;
static char *
db_normalize (void)
{
  // What I'd like here is some kind of a unique string, which would
  // identify the file from anywhere on the network, from Linux or
  // from Win32.  There seems to be no way to do this.  So I err on the
  // side of conservatism:  I strip off the leading path, and I convert
  // everything to lower case.
  char *s, *ss;
  if (!rFilename.Cmp (wxT (""))) // was rFilename=="".
    return ((char *) Name.c_str ());
  Name = rFilename.Lower ();
  for (ss = s = (char *) Name.c_str (); *ss; ss++)
    if (*ss == ':' || *ss == '/' || *ss == '\\')
      s = ss + 1;
  return (s);
}

// Unlock the file.
static void
db_unlock (void)
{
  char *s;
  if (OldPort == 0)
    return;
  if (ReadOnly)
    {
      ReadOnly = 0;
      return;
    }
  //wxMessageBox (db_normalize (), wxT ("Unlock"), wxOK);
  s = db_normalize ();
  if (*s == 0)
    return;
  old_unlock (OldFd, s);
}

// Lock the file.  If Bypass==0, the opportunity to open the file read-only
// is offered.  Otherwise, it is not.  The latter is for things like New
// or SaveAs.
static int
db_trylock (int Bypass)
{
  char *s;
  int ReturnValue;
  if (OldPort == 0)
    return (1);
  if (ReadOnly)
    return (1);
  if (!Bypass)
    {
      ReturnValue =
	wxMessageBox (wxT ("Would you like to open this file read-only?\n"
		      "This is recommended if you don\'t want to edit\n"
		      "the database, since it allows others to edit\n"
		      "and/or view the database.  You\'ll still be\n"
		      "allowed to edit the database in read-only mode\n"
		      "but you\'ll only be able to use \"save-as\" if\n"
		      "you want to save your work."),
		      wxT ("Open read-only?"), wxYES_NO | wxICON_QUESTION);
      if (ReturnValue == wxYES)
	{
	  ReadOnly = 1;
	  return (1);
	}
    }
  //wxMessageBox (db_normalize (), wxT ("Lock"), wxOK);
  s = db_normalize ();
  ReturnValue = old_trylock (OldFd, s);
  if (ReturnValue != 1)
    {
      ReturnValue =
	wxMessageBox (wxT ("The chosen database name cannot be locked.\n"
		      "Probably it is in use, so this operation\n"
		      "cannot be completed.  If you are SURE the\n"
		      "database not in use (perhaps because you\n"
		      "think Do178Builder was previously prematurely\n"
		      "aborted), then I can unlock it for you."),
		      wxT ("Database-lock problem"), wxYES_NO | wxICON_ERROR);
      if (ReturnValue == wxNO)
	return (0);
      ReturnValue = wxMessageBox (wxT ("This is dangerous, since if the database\n"
				  "actually is in use, you may end up overwriting\n"
				  "another developer\'s changes.  Furthermore, if\n"
				  "another user IS using the database and closes\n"
				  "it later, YOUR copy will be unlocked.  It is\n"
				  "best to ask all other users first.  Proceed?"),
				  wxT ("Are you sure?"),
				  wxYES_NO | wxICON_QUESTION);
      if (ReturnValue != wxYES)
	return (0);
      old_unlock (OldFd, s);
      ReturnValue = old_trylock (OldFd, s);
      if (ReturnValue != 1)
	{
	  wxMessageBox (wxT ("I STILL cannot lock the database.  This \n"
			"indicates some deeper problem with the\n"
			"lock-server system.  Sorry!"),
			wxT ("Error!"), wxCANCEL | wxICON_ERROR);
	  return (0);
	}
    }
  return (1);
}

// WDR: class implementations

//----------------------------------------------------------------------------
// SearchBoxClass
//----------------------------------------------------------------------------

// WDR: event table for SearchBoxClass

BEGIN_EVENT_TABLE (SearchBoxClass, wxDialog)
EVT_CHECKBOX (ID_SEARCHTITLES, SearchBoxClass::CheckSearchTitles)
EVT_CHECKBOX (ID_SEARCHDESCRIPTIVE, SearchBoxClass::CheckSearchText)
EVT_CHECKBOX (ID_SEARCHHINT, SearchBoxClass::CheckSearchHints)
EVT_BUTTON (ID_SEARCHNOW, SearchBoxClass::ClickSearchButton)
EVT_BUTTON (ID_SEARCHCANCEL, SearchBoxClass::ClickSearchCancel)
EVT_CHECKBOX (ID_SEARCHALL, SearchBoxClass::CheckAllDocs)
EVT_CHECKBOX (ID_SEARCHPSAC, SearchBoxClass::CheckSearchPsac)
EVT_CHECKBOX (ID_SEARCHSDP, SearchBoxClass::CheckSearchSdp)
EVT_CHECKBOX (ID_SEARCHSVP, SearchBoxClass::CheckSearchSvp)
EVT_CHECKBOX (ID_SEARCHSCMP, SearchBoxClass::CheckSearchScmp)
EVT_CHECKBOX (ID_SEARCHSQAP, SearchBoxClass::CheckSearchSqap)
EVT_CHECKBOX (ID_SEARCHSRS, SearchBoxClass::CheckSearchSrs)
EVT_CHECKBOX (ID_SEARCHSDS, SearchBoxClass::CheckSearchSds)
EVT_CHECKBOX (ID_SEARCHSCS, SearchBoxClass::CheckSearchScs)
EVT_CHECKBOX (ID_SEARCHSRD, SearchBoxClass::CheckSearchSrd)
EVT_CHECKBOX (ID_SEARCHSDD, SearchBoxClass::CheckSearchSdd)
EVT_CHECKBOX (ID_SEARCHSVCP, SearchBoxClass::CheckSearchSvcp)
EVT_CHECKBOX (ID_SEARCHSECI, SearchBoxClass::CheckSearchSeci)
EVT_CHECKBOX (ID_SEARCHSCI, SearchBoxClass::CheckSearchSci)
EVT_CHECKBOX (ID_SEARCHSAS, SearchBoxClass::CheckSearchSas)
END_EVENT_TABLE ();

// WDR: handler implementations for SearchBoxClass

// (This is the position where wxDesigner inserts new code,
// but to keep the source files manageable, I move it later.
// Refer to SearchBoxClass.cpp for the actual code.)


//----------------------------------------------------------------------------
// ProjectOptionEditorClass
//----------------------------------------------------------------------------

// WDR: event table for ProjectOptionEditorClass

BEGIN_EVENT_TABLE (ProjectOptionEditorClass, wxDialog)
EVT_BUTTON (ID_PROJECTOPTIONSOK, ProjectOptionEditorClass::OkProjectOptions)
EVT_BUTTON (ID_PROJECTOPTIONSCANCEL,
	    ProjectOptionEditorClass::CancelProjectOptions)
END_EVENT_TABLE ();

ProjectOptionEditorClass::ProjectOptionEditorClass (wxWindow * parent, wxWindowID id, const wxString & title, const wxPoint & position, const wxSize & size, long style):
wxDialog (parent, id, title, position, size, style)
{
  ProjectOptionEditor (this, TRUE);
}


// WDR: handler implementations for ProjectOptionEditorClass.

void
ProjectOptionEditorClass::CancelProjectOptions (wxCommandEvent & event)
{
  EndModal (0);
}

void
ProjectOptionEditorClass::OkProjectOptions (wxCommandEvent & event)
{
  EndModal (0);
}


//----------------------------------------------------------------------------
// XmlEditorDialogClass
//----------------------------------------------------------------------------

// WDR: event table for XmlEditorDialogClass

BEGIN_EVENT_TABLE (XmlEditorDialogClass, wxDialog)
EVT_TREE_SEL_CHANGED (ID_XMLEDITOR, XmlEditorDialogClass::ClickTree)
EVT_CHECKBOX (ID_PSACCHECK, XmlEditorDialogClass::PsacClick)
EVT_CHECKBOX (ID_SDPCHECK, XmlEditorDialogClass::SdpClick)
EVT_CHECKBOX (ID_SVPCHECK, XmlEditorDialogClass::SvpClick)
EVT_CHECKBOX (ID_SCMPCHECK, XmlEditorDialogClass::ScmpClick)
EVT_CHECKBOX (ID_SQAPCHECK, XmlEditorDialogClass::SqapClick)
EVT_CHECKBOX (ID_SRSCHECK, XmlEditorDialogClass::SrsClick)
EVT_CHECKBOX (ID_SDSCHECK, XmlEditorDialogClass::SdsClick)
EVT_CHECKBOX (ID_SCSCHECK, XmlEditorDialogClass::ScsClick)
EVT_CHECKBOX (ID_SRDCHECK, XmlEditorDialogClass::SrdClick)
EVT_CHECKBOX (ID_DDCHECK, XmlEditorDialogClass::DdClick)
EVT_CHECKBOX (ID_SVCPCHECK, XmlEditorDialogClass::SvcpClick)
EVT_CHECKBOX (ID_SECICHECK, XmlEditorDialogClass::SeciClick)
EVT_CHECKBOX (ID_SCICHECK, XmlEditorDialogClass::SciClick)
EVT_CHECKBOX (ID_SASCHECK, XmlEditorDialogClass::SasClick)
EVT_BUTTON (ID_SATA, XmlEditorDialogClass::ClickSatA)
EVT_BUTTON (ID_SATB, XmlEditorDialogClass::ClickSatB)
EVT_BUTTON (ID_SATC, XmlEditorDialogClass::ClickSatC)
EVT_BUTTON (ID_SATD, XmlEditorDialogClass::ClickSatD)
EVT_BUTTON (ID_SATE, XmlEditorDialogClass::ClickSatE)
EVT_BUTTON (ID_CCA, XmlEditorDialogClass::ClickCcA)
EVT_BUTTON (ID_CCB, XmlEditorDialogClass::ClickCcB)
EVT_BUTTON (ID_CCC, XmlEditorDialogClass::ClickCcC)
EVT_BUTTON (ID_CCD, XmlEditorDialogClass::ClickCcD)
EVT_BUTTON (ID_CCE, XmlEditorDialogClass::ClickCcE)
EVT_BUTTON (ID_ADD_ITEM, XmlEditorDialogClass::AddItem)
EVT_BUTTON (ID_SUBITEM, XmlEditorDialogClass::AddSubItem)
EVT_BUTTON (ID_DELETE_ITEM, XmlEditorDialogClass::DeleteItem)
EVT_BUTTON (ID_SAVEBUTTON, XmlEditorDialogClass::SaveButton)
//EVT_CHECKBOX (ID_SVRCHECK, XmlEditorDialogClass::SvrClick)
//EVT_CHECKBOX (ID_SCMRCHECK, XmlEditorDialogClass::ScmrClick)
//EVT_CHECKBOX (ID_SQARCHECK, XmlEditorDialogClass::SqarClick)
  EVT_BUTTON (ID_FINDBUTTON, XmlEditorDialogClass::FindButton)
EVT_BUTTON (ID_FINDAGAINBUTTON, XmlEditorDialogClass::FindAgainButton)
EVT_BUTTON (ID_MARKUPBUTTON, XmlEditorDialogClass::MarkupButtonClick)
EVT_RADIOBUTTON (ID_NOVICEBUTTON, XmlEditorDialogClass::NoviceCheck)
EVT_RADIOBUTTON (ID_EXPERTBUTTON, XmlEditorDialogClass::ExpertCheck)
EVT_CLOSE (XmlEditorDialogClass::OnClose)
//EVT_SIZE (XmlEditorDialogClass::OnSize) 
  EVT_BUTTON (ID_EMPHASISBUTTON, XmlEditorDialogClass::ClickEmphasisButton)
EVT_BUTTON (ID_COMMANDBUTTON, XmlEditorDialogClass::ClickCommandButton)
EVT_BUTTON (ID_CODEBUTTON, XmlEditorDialogClass::ClickCodeButton)
EVT_BUTTON (ID_LISTINGBUTTON, XmlEditorDialogClass::ClickListingButton)
  //EVT_BUTTON( ID_BUILDBUTTON, XmlEditorDialogClass::BuildButtonClick )
  EVT_BUTTON (ID_DOCKBOOK, XmlEditorDialogClass::DockBookButtonClick)
EVT_BUTTON (ID_TRACEBUTTON, XmlEditorDialogClass::TraceClick)
EVT_BUTTON (ID_DOCBACK, XmlEditorDialogClass::ClickDocBack)
EVT_BUTTON (ID_DOCNEXT, XmlEditorDialogClass::ClickDocNext)
EVT_BUTTON (ID_INSERT_TABLE, XmlEditorDialogClass::ClickInsertTable)
EVT_BUTTON (ID_INSERT_LIST, XmlEditorDialogClass::ClickInsertList)
EVT_BUTTON (ID_INSERT_GRAPHIC, XmlEditorDialogClass::ClickInsertGraphic)
EVT_BUTTON (ID_ADD_ITEM_CUSTOM, XmlEditorDialogClass::AddItemCustomClick)
EVT_BUTTON (ID_SUBITEM_CUSTOM, XmlEditorDialogClass::AddSubitemCustomClick)
EVT_BUTTON (ID_INSERT_HYPERLINK, XmlEditorDialogClass::InsertHyperlinkClick)
EVT_BUTTON (ID_INSERT_FOOTNOTE, XmlEditorDialogClass::InsertFootnoteClick)
EVT_BUTTON (ID_INSERT_MDASH, XmlEditorDialogClass::InsertMdashClick)
EVT_BUTTON (ID_BUILDBUTTON, XmlEditorDialogClass::ClickBuildButton)
END_EVENT_TABLE ();


// WDR: handler implementations for XmlEditorDialogClass.

//------------------------------------------------------------------------------
// Do178BuilderFrame
//------------------------------------------------------------------------------

// WDR: event table for Do178BuilderFrame

BEGIN_EVENT_TABLE (Do178BuilderFrame, wxFrame)
EVT_MENU (ID_ABOUT, Do178BuilderFrame::OnAbout)
EVT_MENU (ID_QUIT, Do178BuilderFrame::OnQuit)
EVT_CLOSE (Do178BuilderFrame::OnCloseWindow)
EVT_MENU (ID_OPEN, Do178BuilderFrame::Open)
EVT_MENU (ID_PROJECTOPTIONS, Do178BuilderFrame::ProjectOptions)
EVT_MENU (ID_EDIT, Do178BuilderFrame::EditDataset)
EVT_MENU (ID_EDITABLEHINTS, Do178BuilderFrame::HintEditCheck)
EVT_MENU (ID_CLOSE, Do178BuilderFrame::CloseFile)
EVT_MENU (ID_SAVE, Do178BuilderFrame::SaveFile)
EVT_MENU (ID_BUILDDOCUMENTS, Do178BuilderFrame::BuildDocuments)
EVT_MENU (ID_NEW, Do178BuilderFrame::OnNew)
EVT_MENU (ID_PRUNE_B, Do178BuilderFrame::PruneB)
EVT_MENU (ID_PRUNE_C, Do178BuilderFrame::PruneC)
EVT_MENU (ID_PRUNE_D, Do178BuilderFrame::PruneD)
EVT_MENU (ID_SAVEAS, Do178BuilderFrame::SaveFileAs) END_EVENT_TABLE ();


Do178BuilderFrame::Do178BuilderFrame (wxWindow * parent,
				      wxWindowID id,
				      const wxString &
				      title,
				      const wxPoint &
				      position,
				      const wxSize & size, long style):
wxFrame (parent, id, title, position, size, style)
{
/*
  if (getenv ("HOME") == NULL)
    getcwd (HomeDirectory, sizeof (HomeDirectory));
  else
    strcpy (HomeDirectory, getenv ("HOME"));
*/

  Centre ();
  Maximize ();

  CreateMyMenuBar ();

  CreateMyToolBar ();

  CreateStatusBar (1);
  SetStatusText (wxT ("Welcome to DO-178 Builder (build " __DATE__ ")"));

  // Splash screen.  Oddly, the wxFatalError doesn't actually show up.
  // But I don't care about that, since it's really all I wanted anyhow.
  if (CopyrightSplash && wxYES !=
      wxMessageBox (wxT ("Do178Builder version " __DATE__ ",\n"
		    "Copyright " COPYRIGHT_YEARS " Ronald S. Burkey.\n"
		    "Do178Builder comes with ABSOLUTELY\n"
		    "NO WARRANTY; for details refer to the\n"
		    "Help menu.  This is free software, and\n"
		    "you are welcome to redistribute it under\n"
		    "certain conditions; refer to the Help\n"
		    "menu for details.  Indicate below if you\n"
		    "agree."),
		    wxT ("Copyright advisory"),
		    wxYES_NO | wxCENTRE | wxICON_INFORMATION))
    //wxFatalError (wxT ("Sorry you feel that way!\nGoodbye until next time."),
    //        wxT ("Abort"));
    wxLogFatalError (wxT ("Sorry you feel that way!\nGoodbye until next time."));

  // insert main window here
}

void
Do178BuilderFrame::LoadXmlToEditor (struct XmlStruct *Xml,
				    XmlEditorDialogClass & dialog,
				    wxTreeItemId Root)
{
  int WasRoot = 0;
  wxTreeItemId NewItem;
  dialog.LastSelected = INVALID_TREE_ITEM;
  if (Root == INVALID_TREE_ITEM)
    {
      WasRoot = 1;
      Root = dialog.ItemTree->AddRoot (SFIX (Xml->Name));
      dialog.ItemTree->SetItemTextColour (Root, *Xml->Color);
      Xml->Id = Root;
      Xml = Xml->Children;
    }
  for (; Xml != NULL; Xml = Xml->Next)
    {
      NewItem = dialog.ItemTree->AppendItem (Root, SFIX (Xml->Name));
      dialog.ItemTree->SetItemTextColour (NewItem, *Xml->Color);
      Xml->Id = NewItem;
      if (Xml->Children != NULL)
	LoadXmlToEditor (Xml->Children, dialog, NewItem);
    }
  if (WasRoot)
    dialog.ItemTree->Expand (Root);
}

void
Do178BuilderFrame::CreateMyMenuBar ()
{
#ifdef __WXMAC__
  wxApp::s_macAboutMenuItemId = ID_ABOUT;
#endif

  SetMenuBar (MyMenuBarFunc ());
}

void
Do178BuilderFrame::CreateMyToolBar ()
{
  wxToolBar *tb = CreateToolBar (wxTB_HORIZONTAL | wxNO_BORDER);

  // tb->AddTool( ID_QUIT, wxBitmap( xxxx_xpm ), wxT ("Quit") );

  tb->Realize ();
}

// WDR: handler implementations for Do178BuilderFrame

void
Do178BuilderFrame::PruneD (wxCommandEvent & event)
{
  if (!FileIsOpen)
    {
      wxMessageBox
	(wxT ("There is no project-dataset file loaded at the moment.\n"
	 "You will have to open a file before pruning it.\n"),
	 wxT ("Sorry!"), wxOK | wxCENTRE | wxICON_EXCLAMATION);
      return;
    }
  if (wxYES !=
      wxMessageBox (wxT ("Pruning to Level D cannot be reversed.  Continue?"),
		    wxT ("Advisory"), wxYES_NO | wxCENTRE | wxICON_WARNING))
    return;
  NormalizeApplicability (&Xml);
  PruneToAssuranceLevel (&Xml, DO178B_LEVEL_D);
  wxMessageBox (wxT ("Pruning to Level D completed."),
		wxT ("Status"), wxCENTRE | wxOK | wxICON_INFORMATION);
}

void
Do178BuilderFrame::PruneC (wxCommandEvent & event)
{
  if (!FileIsOpen)
    {
      wxMessageBox
	(wxT ("There is no project-dataset file loaded at the moment.\n"
	 "You will have to open a file before pruning it.\n"),
	 wxT ("Sorry!"), wxOK | wxCENTRE | wxICON_EXCLAMATION);
      return;
    }
  if (wxYES !=
      wxMessageBox (wxT ("Pruning to Level C cannot be reversed.  Continue?"),
		    wxT ("Advisory"), wxYES_NO | wxCENTRE | wxICON_WARNING))
    return;
  NormalizeApplicability (&Xml);
  PruneToAssuranceLevel (&Xml, DO178B_LEVEL_C);
  wxMessageBox (wxT ("Pruning to Level C completed."),
		wxT ("Status"), wxCENTRE | wxOK | wxICON_INFORMATION);
}

void
Do178BuilderFrame::PruneB (wxCommandEvent & event)
{
  if (!FileIsOpen)
    {
      wxMessageBox
	(wxT ("There is no project-dataset file loaded at the moment.\n"
	 "You will have to open a file before pruning it.\n"),
	 wxT ("Sorry!"), wxOK | wxCENTRE | wxICON_EXCLAMATION);
      return;
    }
  if (wxYES !=
      wxMessageBox (wxT ("Pruning to Level B cannot be reversed.  Continue?"),
		    wxT ("Advisory"), wxYES_NO | wxCENTRE | wxICON_WARNING))
    return;
  NormalizeApplicability (&Xml);
  PruneToAssuranceLevel (&Xml, DO178B_LEVEL_B);
  wxMessageBox (wxT ("Pruning to Level B completed."),
		wxT ("Status"), wxCENTRE | wxOK | wxICON_INFORMATION);
}

void
Do178BuilderFrame::OnNew (wxCommandEvent & event)
{
  int ReturnValue = 0;

  // Optionally save and close any currently-open file.
  if (UnsavedCancel ())
    return;
  FileIsOpen = 0;
  db_unlock ();
  rFilename = wxT ("");

  // Load template file.
  if (0 == (ReturnValue = LoadXmlFile (TemplateFilename, &Xml)))
    {
      rFilename = wxT ("Untitled.xml");
      SetStatusText (rFilename);
      FileIsOpen = 1;
      // We don't actually try to obtain a lock on Untitled.xml.  Anyone silly
      // enough to save stuff under this name deserves to get burned.

      // Save under some better name.  
      wxString NewFilename;
      NewFilename =
	wxFileSelector (wxT ("Save dataset under new filename"), wxT ("."), rFilename,
			wxT ("xml"), wxT ("*.xml"), wxSAVE | wxOVERWRITE_PROMPT);
      if (NewFilename)
	{
	  SetStatusText (NewFilename);
	  rFilename = NewFilename;
	  if (1 != db_trylock (1))
	    {
	      //wxMessageBox (wxT ("Could not lock database.\n"
	      //        "Perhaps it is in use.\n"
	      //  	"Retaining the name Untitled.xml."), 
	      //        wxT ("Error"), wxCENTRE | wxOK | wxICON_ERROR);
	      rFilename = wxT ("Untitled.xml");
	      SetStatusText (rFilename);
	    }
	  else
	    SaveFile (event);
	}

      // And edit it!
      EditDataset (event);
    }
  else
    {
      sprintf (s, "Could not load %s.\nError code #%d", TemplateFilename,
	       ReturnValue);
      wxMessageBox (SFIX (s), wxT ("Error"), wxCENTRE | wxOK | wxICON_ERROR);
    }

}

// Returns 0 on success, 1 on error.
int
BuildTheDocuments (int BatchRun)
{
  //char s[1024]; 
  char *ss;
  wxString OutputDirectory, SourceDir;
  struct stat StatBuf;
  int i, RetVal;
  static char sOD[SOD_SIZE];	// A dummy variable.


  // Figure out the name of the output directory.  We do this
  // by taking the name of the source file and creating a 
  // subdirectory with the same name (but without the extension
  // ".documents" instead of ".xml".
  strcpy (s, rFilename.char_str ());
  ss = strstr (s, ".xml");
  if (ss == NULL || 0 != strcmp (ss, ".xml"))
    {
      sprintf (s, "Your dataset filename does not end in \".xml\".");
      goto ErrorExit;
    }
  if (OutputOverride)
    strcpy (ss, ".documents");
  else
    {
      time_t t;
      struct tm *ts;
      time (&t);
      ts = localtime (&t);
      sprintf (ss, "-%04d%02d%02d.documents",
	       ts->tm_year + 1900, ts->tm_mon + 1, ts->tm_mday);
    }
  OutputDirectory = SFIX (s);

  // Create the output directory, if necessary.
  strcpy (sOD, OutputDirectory.char_str ());
  i = stat (sOD, &StatBuf);
  if (i != 0)
    {
      // Output directory doesn't exist (or its name is bollixed.
      if (0 != MKDIR (sOD))
	{
	  sprintf (s, "Cannot create the ouput directory\n\"%s\"", sOD);
	  goto ErrorExit;
	}
    }
  else if ((StatBuf.st_mode & S_IFDIR) == 0)
    {
      // The output directory exists -- except that it ISN'T a directory.
      sprintf (s, "The name we want to use for the output directory\n"
	       "is not available:\n%s", sOD);
      goto ErrorExit;
    }
  else if (!BatchRun)
    {
      // The output directory really exists.
      sprintf (s, "The output directory already exists:\n"
	       "%s\nIs it okay to overwrite the existing contents?",
	       sOD);
      if (wxYES !=
	  wxMessageBox (SFIX (s), wxT ("Advisory"), wxYES_NO | wxCENTRE | wxICON_WARNING))
	{
	  sprintf (s, "Aborted by user.");
	  goto ErrorExit;
	}
    }
  OutputDirectory = OutputDirectory + wxT ("/");

  // Create the documents themselves.  
  strcpy (sOD, OutputDirectory.char_str ());
  RetVal = XmlToSgml (&Xml, sOD);
  if (RetVal == 0)
    {
      const char *Extension;
      i = strlen (sOD);
      if (i < 30)
	i = (40 - i) / 2;
      else
	i = 0;
      if (x4)
	Extension = "xml";
      else
	Extension = "sgml";
      if (Do254)
	sprintf (s, "The DO-254 documents have been created, in\n"
		 "%*s%s.\n"
		 "They are DocBook files with names like PHAC.%s,\n"
		 "HDP.%s, HVVP.%s, HCMP.%s, and so on.",
		 i, "", sOD, Extension, Extension,
		 Extension, Extension);
      else
	sprintf (s, "The DO-178B documents have been created, in\n"
		 "%*s%s.\n"
		 "They are DocBook files with names like PSAC.%s,\n"
		 "SDP.%s, SVP.%s, SCMP.%s, and so on.",
		 i, "", sOD, Extension, Extension,
		 Extension, Extension);
      if (BatchRun)
        printf ("%s\n", s);
      else
        wxMessageBox (SFIX (s), wxT ("Status"), wxCENTRE | wxOK | wxICON_INFORMATION);
      return (0);
    }
  else
    {
      sprintf (s,
	       "Could not create documents.\nError code #%d on document #%d",
	       RetVal & 0xff, (RetVal >> 8) & 0xff);
    ErrorExit:
      if (BatchRun)
        printf ("%s\n", s);
      else
        wxMessageBox (SFIX (s), wxT ("Error"), wxCENTRE | wxOK | wxICON_ERROR);
      return (1);
    }
}

void
Do178BuilderFrame::BuildDocuments (wxCommandEvent & event)
{
  if (Do254)
    SetStatusText (wxT ("Building DO-254 documents -- please wait --."));
  else
    SetStatusText (wxT ("Building DO-178B documents -- please wait ..."));
  if (BuildTheDocuments (0))
    SetStatusText (wxT ("Errors occurred in the document-build."));
  else
    SetStatusText (wxT ("Document-build completed."));
}

void
Do178BuilderFrame::SaveFileAs (wxCommandEvent & event)
{
  wxString NewFilename, OldFilename;
  int OldFileIsOpen, OldReadOnly;
  OldFilename = rFilename;
  OldFileIsOpen = FileIsOpen;
  OldReadOnly = ReadOnly;
  FileIsOpen = 0;
  db_unlock ();
  NewFilename =
    wxFileSelector (wxT ("Save dataset under new filename"), wxT ("."), rFilename, wxT ("xml"),
		    wxT ("*.xml"), wxSAVE | wxOVERWRITE_PROMPT);
  if (NewFilename)
    {
      rFilename = NewFilename;
      if (1 != db_trylock (1))
	goto Aborted;
      SetStatusText (NewFilename);
      SaveFile (event);
    }
  else
    {
    Aborted:
      FileIsOpen = OldFileIsOpen;
      rFilename = OldFilename;
      ReadOnly = OldReadOnly;
      db_trylock (1);
    }
}

void
Do178BuilderFrame::SaveFile (wxCommandEvent & event)
{
  static char sFN[SOD_SIZE];
  strcpy (sFN, rFilename.char_str ());
  if (ReadOnly)
    {
      wxMessageBox (wxT ("This file was opened read-only.\n"
		    "Please use /File/SaveAs instead."), wxT ("Error"),
		    wxCENTRE | wxOK | wxICON_ERROR);
    }
  else if (SaveXmlFile (sFN, &Xml))
    wxMessageBox (wxT ("Could not save the file."), wxT ("Error"),
		  wxCENTRE | wxOK | wxICON_ERROR);
  else
    DatasetChanged = 0;
}

void
Do178BuilderFrame::CloseFile (wxCommandEvent & event)
{
  if (UnsavedCancel ())
    return;
  FileIsOpen = 0;
  db_unlock ();
  rFilename = wxT ("");
}

void
Do178BuilderFrame::HintEditCheck (wxCommandEvent & event)
{
  //HintEditable = !HintEditable;
  //wxObject *Object = event.GetEventObject ();
  wxMenuBar *MenuBar = GetMenuBar ();
  wxMenuItem *MenuItem = MenuBar->FindItem (ID_EDITABLEHINTS);
  HintEditable = MenuItem->IsChecked ();
  if (HintEditable && wxYES != wxMessageBox
      (wxT ("Framework-editing mode is an administrative function,\n"
       "not needed in a typical design project.  It can\n"
       "introduce incompatibilities with existing project-\n"
       "datasets, if while in framework-editing mode you add,\n"
       "delete, or rename data items.  On the other hand,\n"
       "it is perfectly safe to use it for editing HINTS.\n"
       "Are you sure you want to proceed?"),
       wxT ("Compatibility danger!"), wxYES_NO | wxCENTRE | wxICON_EXCLAMATION))
    {
      HintEditable = FALSE;
      MenuItem->Check (FALSE);
    }
}

void
Do178BuilderFrame::EditDataset (wxCommandEvent & event)
{
  //wxFont Font (12, wxMODERN, wxNORMAL, wxNORMAL);
  int x, y, width, height;
#ifdef WIN32
  // The following is a workaround for the fact that I can't seem to get
  // wxClientDisplayRect to work right (at compile time) in Win32 with
  // wxGTK 2.7, and that I can't get wxGTK 2.4.2 to work at all in Win32.
  x = y = 0;
  width = MaxWindowX;
  height = MaxWindowY;
#else // WIN32
  wxClientDisplayRect (&x, &y, &width, &height);
  if (width > MaxWindowX)
    width = MaxWindowX;
  if (height > MaxWindowY)
    height = MaxWindowY;
#endif
  wxPoint Point (x, y);
  wxSize Size (width, height);
  printf ("%d %d %d %d\n", Point.x, Point.y, Size.GetWidth (),
	  Size.GetHeight ());
  if (!FileIsOpen)
    {
      wxMessageBox
	(wxT ("There is no project-dataset file loaded at the moment.\n"
	 "You will have to open a file before editing it.\n"),
	 wxT ("Sorry!"), wxOK | wxCENTRE | wxICON_EXCLAMATION);
      return;
    }
  const char *DialogTitle;
  if (Do254)
    DialogTitle = "DO-254 documentation dataset editor";
  else
    DialogTitle = "DO-178B documentation dataset editor";
  long ExtraStyle;
  if (ResizeWorkaround)
    ExtraStyle = 0;
  else
    ExtraStyle = wxRESIZE_BORDER | wxMAXIMIZE_BOX;
  XmlEditorDialogClass dialog (this, -1, SFIX (DialogTitle), wxDefaultPosition, Size,	//wxDefaultPosition, wxDefaultSize,
			       wxDEFAULT_DIALOG_STYLE | wxSYSTEM_MENU |
			       ExtraStyle);
  dialog.RootXml = &Xml;
  dialog.Template = &TemplateXml;
  // Now load the actual project data into the dialog.
  LoadXmlToEditor (&Xml, dialog, INVALID_TREE_ITEM);
  dialog.ItemTree->SelectItem (Xml.Id);
  wxTreeEvent DummyEvent;
  dialog.ClickTreeBypass = 1;
  dialog.ClickTree (DummyEvent);
  //dialog.SetFont (Font);
  dialog.SetSize (width, height);
  dialog.Center ();
  dialog.ShowModal ();
  SetStatusText (rFilename);
}

void
Do178BuilderFrame::ProjectOptions (wxCommandEvent & event)
{
  ProjectOptionEditorClass dialog (this, -1, wxT ("Project options"));
  dialog.Center ();
  dialog.ShowModal ();
}

int
Do178BuilderFrame::UnsavedCancel (void)
{
  static char sFN[SOD_SIZE];
  strcpy (sFN, rFilename.char_str ());
  if (DatasetChanged)
    switch (wxMessageBox
	    (wxT ("The dataset has changed.  These changes will be lost.\n"
	     "Save the dataset before continuing?"), wxT ("Confirm"),
	     wxYES_NO | wxCANCEL | wxCENTRE | wxICON_QUESTION))
      {
      case wxYES:
	if (ReadOnly)
	  {
	    wxMessageBox (wxT ("This file was opened read-only.\n"
			  "It can only be saved with SaveAs."),
			  wxT ("Error"), wxCENTRE | wxOK | wxICON_ERROR);
	    return (1);
	  }
	if (SaveXmlFile (sFN, &Xml))
	  {
	    wxMessageBox (wxT ("Cannot save the file.\nPerhaps the disk is full."),
			  wxT ("Error"), wxCENTRE | wxOK | wxICON_ERROR);
	    return (1);
	  }
	break;
      case wxNO:
	break;
      case wxCANCEL:
      default:
	return (1);
      }
  DatasetChanged = 0;
  return (0);
}

void
Do178BuilderFrame::Open (wxCommandEvent & event)
{
  static char sFN[SOD_SIZE];
  int ReturnValue;
  //char s[128];
  if (UnsavedCancel ())
    return;
  FileIsOpen = 0;
  db_unlock ();
  rFilename =
    wxFileSelector (wxT ("Open existing dataset"), SFIX (CurrentDirectory), wxT (""), wxT ("xml"),
		    wxT ("*.xml"), wxOPEN | wxFILE_MUST_EXIST);
  if (!rFilename.IsEmpty ())
    {
      strcpy (sFN, rFilename.char_str ());
      if (1 != db_trylock (0))
	{
	  //wxMessageBox (wxT ("Cannot lock the database;\nperhaps it is in use already."),
	  //wxT ("Error"), wxCENTRE | wxOK | wxICON_ERROR); 
	  rFilename = wxT ("");
	  strcpy (sFN, "");
	}
      else if (0 == (ReturnValue = LoadXmlFile (sFN, &Xml)))
	{
	  SetStatusText (rFilename);
	  FileIsOpen = 1;
	  EditDataset (event);
	}
      else
	{
	  extern long LowLevelError;
	  extern int LowReturnCode;
	  if (LowLevelError != -1)
	    {
	      sprintf (s, "Could not load the file.\nError code #%d@%ld",
		       LowReturnCode, LowLevelError);
	      LowLevelError = -1;
	    }
	  else
	    sprintf (s, "Could not load the file.\nError code #%d",
		     ReturnValue);
	  wxMessageBox (SFIX (s), wxT ("Error"), wxCENTRE | wxOK | wxICON_ERROR);
	  db_unlock ();
	  rFilename = wxT ("");
	}
    }
}

void
Do178BuilderFrame::OpenTemplate (void)
{
  int ReturnValue;
  //char s[1024];
  char s[60];
  static char sFN[SOD_SIZE];
  
  sprintf (s, "%s/%s", HomeDirectory, TemplateFilename);
/*  
  char *ss, *sss;
  strcpy (s, argv[0]);
  for (sss = NULL, ss = s; *ss; ss++)
    if (*ss == '/')
      sss = ss;
  if (sss != NULL)
    {    
      *sss = '\0';
      strcpy (HomeDirectory, s);
      strcat (s, "/");    
      strcat (s, TemplateFilename);    
    }
  else
    strcpy (s, TemplateFilename);  
*/
  ReturnValue = LoadXmlFile (s, &TemplateXml);
  if (0 == ReturnValue)
    return;
  sprintf (s, "Please locate hidden file %s", TemplateFilename);
  rFilename =
    wxFileSelector (SFIX (s),
		    SFIX (HomeDirectory), wxT (""), wxT ("xml"), SFIX (TemplateFilename),
		    wxOPEN | wxFILE_MUST_EXIST);
  strcpy (sFN, rFilename.char_str ());
  if (0 == (ReturnValue = LoadXmlFile (sFN, &TemplateXml)))
    return;
  else
    {
      sprintf (s, "Could not load %s.\nError code #%d",
	       TemplateFilename, ReturnValue);
      wxMessageBox (SFIX (s), wxT ("Fatal error"), wxCENTRE | wxOK | wxICON_ERROR);
      close (TRUE);
    }
}

void
Do178BuilderFrame::OnAbout (wxCommandEvent & event)
{
  wxMessageDialog dialog (this,
			  wxT ("Welcome to DO-178 Builder (build " __DATE__
			  ")\nCopyright " COPYRIGHT_YEARS " Ron Burkey"),
			  wxT ("About DO-178 Builder"), wxOK | wxICON_INFORMATION);
  dialog.Center ();
  dialog.ShowModal ();
}

void
Do178BuilderFrame::OnQuit (wxCommandEvent & event)
{
  if (UnsavedCancel ())
    return;
  Close (TRUE);
  db_unlock ();
  rFilename = wxT ("");
  FileIsOpen = 0;
}

void
Do178BuilderFrame::OnCloseWindow (wxCloseEvent & event)
{
  // if ! saved changes -> return
  if (UnsavedCancel ())
    return;
  Destroy ();
  db_unlock ();
  rFilename = wxT ("");
  FileIsOpen = 0;
}

//------------------------------------------------------------------------------
// Do178BuilderApp
//------------------------------------------------------------------------------

IMPLEMENT_APP (Do178BuilderApp) Do178BuilderApp::Do178BuilderApp ()
{
}

bool Do178BuilderApp::OnInit ()
{
  int
    m,
    n;
  int
    FileIsOpen = 0;
  int
    Batch = 0;
  static char sOD[SOD_SIZE];

  // We figure out the home directory by examining the path of the program
  // name.  If none, we use the current working directory.
  //char s[1024]; 
  char *
  ss, *
    sss;
  strcpy (s, SFIX (argv[0]).char_str ());
  printf ("%s\n", s);
  for (sss = NULL, ss = s; *ss; ss++)
    if (*ss == '/')
      sss = ss;
  if (sss == NULL)
    {
      getcwd (s, sizeof (s));
      ss = s + strlen (s) - 1;
      if (*ss == '/')
	*ss = '\0';
    }
  else
    *sss = '\0';
  strcpy (HomeDirectory, s);

  // Examine the command-line arguments.  
  int
    i;
    
  for (i = 1; i < argc; i++)
    {
      wxString wxs (argv[i]);
      strcpy (sOD, wxs.char_str ());
      if (0 == strcasecmp (sOD, "-h")
	  || 0 == strcasecmp (sOD, "--help")
	  || 0 == strcasecmp (sOD, "-?"))
	{
	Help:
	  printf ("---------------------------------------");
	  printf ("---------------------------------------\n");
	  printf ("Do178Builder, build %s\n", __DATE__);
	  printf ("Copyright " COPYRIGHT_YEARS " by Ronald S. Burkey.\n");
	  printf ("Licensed under the GPL.\n");
	  printf ("\n");
	  printf ("Do178Builder helps incorporate DO-178B into the \n");
	  printf ("software-development effort.  Specifically, it creates\n");
	  printf ("DO-178B documents for aviation-certification purposes.\n");
	  printf ("As of 2006-06-23, it can also help to create DO-254\n");
	  printf ("life-cycle data.\n");
	  printf ("\n");
	  printf ("USAGE:\tDo178Builder [OPTIONS] [XMLfile]\n");
	  printf ("\n");
	  printf ("OPTIONS:\n");
	  printf ("\t-h           Display this help information.\n");
	  printf ("\t-p=dirname   Specify your project directory.\n");
	  printf ("\t-c           Eliminate the copyright splashscreen\n");
	  printf ("\t-o           By default, the SGML output files are\n");
	  printf ("\t             placed in a directory named according\n");
	  printf ("\t             to the project and current date. With\n");
	  printf
	    ("\t             the -o switch, this behavior is overridden\n");
	  printf ("\t             and only the project name is used.\n");
	  printf
	    ("\t--batch      Bypasses the Do178Builder GUI, and simply\n");
	  printf
	    ("\t             builds the DocBook output files from the \n");
	  printf
	    ("\t             specified Do178Builder database and exits.\n");
	  printf ("\t--x412, --x42, --x43, --x44, --x45\n");
	  printf
	    ("\t             By default, Do178Builder outputs SGML files\n");
	  printf
	    ("\t             compatible with DocBook 3.1.  These command-\n");
	  printf
	    ("\t             line switches instead select DocBook XML, \n");
	  printf
	    ("\t             versions 4.1.2, 4.2, 4.3, 4.4, or 4.5.  Note,\n");
	  printf
	    ("\t             however, that this may work imperfectly if\n");
	  printf
	    ("\t             DocBook 3.1 commands are embedded directly\n");
	  printf
	    ("\t             within the body of the text.  Do178Builder\n");
	  printf
	    ("\t             attempts to detect and convert the DocBook\n");
	  printf
	    ("\t             commands it finds there, but your ingenuity\n");
	  printf
	    ("\t             in creating the DocBook commands is greater\n");
	  printf
	    ("\t             than Do178Builder\'s ingenuity in correcting\n");
	  printf ("\t             them, so there may still be some manual\n");
	  printf ("\t             corrections required.\n");
	  printf
	    ("\t--combine-srd-sdd   By default, the SRD and SDD documents\n");
	  printf
	    ("\t             are separate, but with this switch the two\n");
	  printf ("\t             are automatically merged into a single \n");
	  printf
	    ("\t             document when the output SGML is created.\n");
	  printf
	    ("\t             (There is no effect whatever on your XML.)\n");
	  printf
	    ("\t             This can be accomplished also by editing\n");
	  printf ("\t             the document framework, but this makes\n");
	  printf
	    ("\t             much more simple to do and to undo.  If you\n");
	  printf
	    ("\t             can get away with a single merged SRD/SDD,\n");
	  printf
	    ("\t             it can ease up some traceability issues.\n");
	  printf ("\t--combine-plans=P  By default, the DO-178B planning docs\n");
	  printf ("\t             (or their DO-254 equivalent documents)\n");
	  printf ("\t             are separate outputs, but with this\n");
	  printf ("\t             switch they are merged into a single \n");
	  printf ("\t             document.  The single output plan can be\n");
	  printf ("\t             named as if it were any of the plans, and\n");
	  printf ("\t             P selects which naming scheme is used.  P\n");
	  printf ("\t             must be one of the following:  PSAC, SDP,\n");
	  printf ("\t             SVP, SQAP, SCMP, PHAC, HDP, HVVP, HPAP, HCMP.\n");
	  printf ("\t--no-forward-trace  By default, when documents are\n");
	  printf ("\t             built, both backward and forward\n");
	  printf
	    ("\t             traceability references are added to the\n");
	  printf ("\t             documents.  For example, the SDD traces\n");
	  printf
	    ("\t             to the SRD, but the SRD also traces to the\n");
	  printf
	    ("\t             SDD.  Forward traceability is problematic\n");
	  printf
	    ("\t             in the minds of some DERs for a couple of\n");
	  printf
	    ("\t             reasons:  a) One of the documents doesn\'t\n");
	  printf ("\t             exist yet, so how can you know that the\n");
	  printf
	    ("\t             trace references are correct?  And, b) forward\n");
	  printf
	    ("\t             traceability is usually one-to-several (e.g.,\n");
	  printf
	    ("\t             any given section in the SDD traces to a\n");
	  printf
	    ("\t             single section of the SRD, but any given\n");
	  printf
	    ("\t             SRD section of the SRD probably traces to\n");
	  printf
	    ("\t             several SDD sections, so how can you know that\n");
	  printf
	    ("\t             the trace is complete?  For this reason, you\n");
	  printf
	    ("\t             may want to simply eliminate forward trace\n");
	  printf
	    ("\t             references.  The --no-forward-trace command-\n");
	  printf ("\t             line switch accomplishes this.\n");
	  printf
	    ("\t--no-productname  By default, the DocBook output produced\n");
	  printf
	    ("\t             by Do178Builder encapsulates your product name\n");
	  printf
	    ("\t             in the DocBook tags <productname></productname>\n");
	  printf
	    ("\t             on the title page and in page headers.  One side-\n");
	  printf
	    ("\t             effect of that is that with some common style-\n");
	  printf
	    ("\t             sheets the product name is followed by a TM\n");
	  printf
	    ("\t             (trademark) symbol.  I personally find this \n");
	  printf
	    ("\t             annoying and besides, it's easy enough to add\n");
	  printf
	    ("\t             \"&trade;\" manually if you like.  While this\n");
	  printf
	    ("\t             could be changed in the stylesheet, alternately\n");
	  printf
	    ("\t             you can use the --no-productname switch to simply\n");
	  printf
	    ("\t             eliminate the tags completely.  On the other hand,\n");
	  printf
	    ("\t             with some stylesheets you are not *allowed* to \n");
	  printf
	    ("\t             do without the <productname> tag.  Your mileage\n");
	  printf ("\t             may vary.\n");
	  printf ("\t--no-resize-workaround  On some platforms, there is\n");
	  printf ("\t             a problem which causes the program\n");
	  printf ("\t             to freeze up at some point after the \n");
	  printf
	    ("\t             editing dialog has been resized larger than\n");
	  printf
	    ("\t             1280x1024 pixels.  This command-line switch\n");
	  printf
	    ("\t             deactivates the workaround for that problem.\n");
	  printf ("\t             The workaround does appear to work, but\n");
	  printf
	    ("\t             it\'s irritating if activated unnecessarily\n");
	  printf ("\t             because window resizing is limited to \n");
	  printf
	    ("\t             1280x1024 and is inconsistent in several\n");
	  printf
	    ("\t             ways.  But if you deactivate it, you risk\n");
	  printf
	    ("\t             losing work when the program suddenly exits.\n");
	  printf
	    ("\t--no-trace   Some DERs simply spit up on the whole approach\n");
	  printf
	    ("\t             Do178Builder takes to traceability.  This that\'s\n");
	  printf
	    ("\t             the case, you can completely turn off trace\n");
	  printf
	    ("\t             references with this command-line switch.\n");
	  printf ("\t-x=N         Overrides maximum horizontal size\n");
	  printf ("\t             (in pixels) of the editing window.\n");
	  printf
	    ("\t             Useful only without --no-resize-workaround.\n");
	  printf ("\t             Default is %d.\n", MaxWindowX);
	  printf ("\t-y=N         Same for vertical size.  Default=%d.\n",
		  MaxWindowY);
	  printf ("\t-s=ipaddress IP address or network name of the\n");
	  printf ("\t             computer on which the server for file-\n");
	  printf ("\t             locking is running.  Currently \"%s\".\n",
		  OldServer);
	  printf ("\t             Presently not available for Win32.\n");
	  printf ("\t-l=portnum   Port-number used for accessing the lock\n");
	  printf ("\t             server.  Currently %d.  You can bypass\n",
		  OldPort);
	  printf ("\t             the lock-server completely with -l=0.\n");
	  printf ("\t-f=f[,f]     Changes the font-size used for the item\n");
	  printf
	    ("\t             text window.  The font will be bold-face\n");
	  printf
	    ("\t             if the number is positive, non-bold if it\n");
	  printf
	    ("\t             is negative.  If the optional second number\n");
	  printf
	    ("\t             is present, it is used for the item-tree window.\n");
	  printf
	    ("\t             Otherwise, the same size is used for both,\n");
	  printf
	    ("\t             except that the item-tree window will be\n");
	  printf
	    ("\t             normal weight rather than bold.  The default\n");
	  printf ("\t             is -f=%d,%d.\n", FontSize, TreeFontSize);
	  printf
	    ("\t--DO-254     Work with DO-254 rather than DO-178B.  The XML\n");
	  printf ("\t             database is consistent between the two,\n");
	  printf ("\t             so DO-178B docs can be imported into\n");
	  printf ("\t             DO-254, and vice-versa.\n");
	  printf ("\n");
	  printf
	    ("If the optional XMLfile is specified, it will be loaded.\n");
	  printf ("---------------------------------------");
	  printf ("---------------------------------------\n");
	  exit (1);
	}
      else if (!strcmp (sOD, "--DO-254"))
	{
	  Do254 = 1;
	  TemplateFilename = ".template254.xml";
	}
      else if (!strcmp (sOD, "--no-productname"))
	NoProductname = 1;
      else if (!strcmp (sOD, "--no-resize-workaround"))
	ResizeWorkaround = 0;
      else if (!strcmp (sOD, "--no-forward-trace"))
	NoForwardTrace = 1;
      else if (!strcmp (sOD, "--no-trace"))
	NoTrace = 1;
      else if (!strcmp (sOD, "--combine-srd-sdd"))
	CombineSrdSdd = 1;
      else if (!strncmp (sOD, "--combine-plans=", 16))
        CombinedPlanName = &sOD[16];
      else if (!strcmp (sOD, "--batch"))
        Batch = 1;
      else if (!strcmp (sOD, "--x412"))
	{
	  x4 = 1;
	  x4v1 = "4.1.2";
	  x4v2 = "4.0";
	}
      else if (!strcmp (sOD, "--x42"))
	{
	  x4 = 1;
	  x4v1 = "4.2";
	  x4v2 = x4v1;
	}
      else if (!strcmp (sOD, "--x43"))
	{
	  x4 = 1;
	  x4v1 = "4.3";
	  x4v2 = "4.3b2";
	}
      else if (!strcmp (sOD, "--x44"))
	{
	  x4 = 1;
	  x4v1 = "4.4";
	  x4v2 = x4v1;
	}
      else if (!strcmp (sOD, "--x45"))
	{
	  x4 = 1;
	  x4v1 = "4.5";
	  x4v2 = x4v1;
	}
      else if (1 == sscanf (sOD, "-x=%d", &n))
	MaxWindowX = n;
      else if (1 == sscanf (sOD, "-y=%d", &n))
	MaxWindowY = n;
      else if (1 == sscanf (sOD, "-l=%d", &n))
	OldPort = n;
      else if (0 == strcmp (sOD, "-o"))
	OutputOverride = 1;
      else if (0 == strncmp (sOD, "-s=", 3))
	OldServer = &sOD[3];
      else if (0 == strncasecmp (sOD, "-p=", 3))
	{
	  strcpy (CurrentDirectory, &sOD[3]);
	  chdir (CurrentDirectory);
	  strcpy (HomeDirectory, CurrentDirectory);
	}
      else if (0 == strcasecmp (sOD, "-c"))
	CopyrightSplash = 0;
      else if (2 == sscanf (sOD, "-f=%d,%d", &n, &m))
	{
	  FontSize = n;
	  TreeFontSize = m;
	}
      else if (1 == sscanf (sOD, "-f=%d", &n))
	{
	  FontSize = TreeFontSize = n;
	  if (n > 0)
	    TreeFontSize = -n;
	}
      else if (!FileIsOpen)
	{
	  if (0 == LoadXmlFile (sOD, &Xml))
	    {
	      nFilename = argv[i];
	      FileIsOpen = 1;
	    }
	}
      else
	{
	  printf ("Unknown argument:  \"%s\"\n", sOD);
	  goto Help;
	}
    }
    
  // Sanity check on --combine-plans=CombinedPlanName
  if (CombinedPlanName != NULL)
    {
      if (Do254)
	{
          if (strcmp (CombinedPlanName, "PHAC") &&
	      strcmp (CombinedPlanName, "HDP") &&
	      strcmp (CombinedPlanName, "HVVP") &&
	      strcmp (CombinedPlanName, "HPAP") &&
	      strcmp (CombinedPlanName, "HCMP"))
	    {
	      printf ("Not a legal plan name: --combine-plans=\"%s\"\n.", CombinedPlanName);
	      exit (1);
	    }
	}
      else
	{
          if (strcmp (CombinedPlanName, "PSAC") &&
	      strcmp (CombinedPlanName, "SDP") &&
	      strcmp (CombinedPlanName, "SVP") &&
	      strcmp (CombinedPlanName, "SQAP") &&
	      strcmp (CombinedPlanName, "SCMP"))
	    {
	      printf ("Not a legal plan name: --combine-plans=\"%s\"\n.", CombinedPlanName);
	      exit (1);
	    }
	}
    }

  // If batch processing and bypassing GUI, take care of it now before
  // proceeding with GUI setup.
  if (Batch)
	{
      if (!FileIsOpen)
	    {
	  fprintf (stderr, "Do178Builder database not specified or not found.\n");
	  return (1);
	}
      rFilename = nFilename;
      if (BuildTheDocuments (1))
	    {
	  fprintf (stderr, "There was an error building the documents.\n");
	  return (1);
	}
      fprintf (stderr, "Success!\n");
      return (0);
    }

  // Okay, continue with normal setup.
  const char *FrameTitle;
  if (Do254)
    FrameTitle = "DO-254 Builder";
  else
    FrameTitle = "DO-178B Builder";
  Do178BuilderFrame *
    frame = new Do178BuilderFrame (NULL, -1, SFIX (FrameTitle), wxPoint (20, 20),
			   wxSize (500, 340));
  //frame->SetFont( wxFont( FontSize, wxSWISS, wxNORMAL, wxNORMAL ) );
  frame->OpenTemplate ();
  frame->Maximize (FALSE);
  frame->Show (TRUE);

  // Let's start talking to the lock server.
  if (OldPort != 0)
    {
      OldFd = old_connect ((char *) OldServer, OldPort);
      if (OldFd == -1)
	{
	  wxMessageBox (wxT ("Cannot connect to the lock server."),
			wxT ("Fatal Error"), wxOK | wxICON_ERROR);
	  frame->Destroy ();
	  return (TRUE);
	}
    }

  if (FileIsOpen)
    {
      rFilename = nFilename;

      // We need to get a lock on the file.
      if (1 != db_trylock (0))
	{
	  //wxMessageBox (wxT ("Cannot lock the database; it may already be in use."),
	  //  wxT ("Fatal Error"), wxOK | wxICON_ERROR);
	  rFilename = wxT ("");
	  //frame->Destroy ();
	  return (TRUE);
	}

      frame->FileIsOpen = FileIsOpen;
      frame->SetStatusText (nFilename);
      /*
         At this point, we presumably need to use wxPostEvent to set a
         pending menu click on the ID_EDIT menu item that will later 
         activate the frame->EditDataset function.  This function
         cannot be called directly at this point.  Unfortunately,
         I don't yet know how to make that happen.
       */
    }

  return TRUE;
}

int
Do178BuilderApp::OnExit ()
{
  db_unlock ();
  rFilename = wxT ("");
  return 0;
}
